#!/usr/bin/env python2
# -*- coding: utf-8 -*-

#    Copyright 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import json
import os
import sys
import urllib2
from timmy import tools
from timmy import conf
from timmy.nodes import NodeManager as BaseNodeManager
from timmy.nodes import Node as BaseNode

try:
    import fuelclient
    if hasattr(fuelclient, 'connect'):
        # fuel > 9.0.1 - drop support, use API and CLI instead
        FuelClient = None
    else:
        import fuelclient.client
        if type(fuelclient.client.APIClient) is fuelclient.client.Client:
            # fuel 9.0.1 and below
            from fuelclient.client import Client as FuelClient
        else:
            FuelClient = None
except:
    FuelClient = None

try:
    from fuelclient.client import logger
    logger.handlers = []
except:
    pass


def add_args(parser):
    parser.add_argument('--fuel-ip', help='fuel ip address')
    parser.add_argument('--fuel-user', help='fuel username')
    parser.add_argument('--fuel-pass', help='fuel password')
    parser.add_argument('--fuel-token', help='fuel auth token')
    parser.add_argument('--fuel-logs-no-remote', action='store_true',
                        help='Do not collect remote logs from Fuel.')
    parser.add_argument('--fuel-proxy',
                        help='use os system proxy variables for fuelclient',
                        action='store_true')
    parser.add_argument('-j', '--nodes-json',
                        help=('Path to a json file retrieved via'
                              ' "fuel node --json". Useful to speed up'
                              ' initialization, skips "fuel node" call.'))
    return parser


def check_args(args, conf):
    if args.fuel_ip:
        conf['fuel_ip'] = args.fuel_ip
    if args.fuel_user:
        conf['fuel_user'] = args.fuel_user
    if args.fuel_pass:
        conf['fuel_pass'] = args.fuel_pass
    if args.fuel_proxy:
        conf['fuel_skip_proxy'] = False
    if args.fuel_token:
        conf['fuel_api_token'] = args.fuel_token
        conf['fuelclient'] = False
    if args.fuel_logs_no_remote:
        conf['fuel_logs_no_remote'] = True


def add_conf(conf):
    conf['fuel_ip'] = '127.0.0.1'
    conf['fuel_api_user'] = 'admin'
    conf['fuel_api_pass'] = 'admin'
    conf['fuel_api_token'] = None
    conf['fuel_api_tenant'] = 'admin'
    conf['fuel_api_port'] = '8000'
    conf['fuel_api_keystone_port'] = '5000'
    # The three parameters below are used to override FuelClient, API, CLI auth
    conf['fuel_user'] = None
    conf['fuel_pass'] = None
    conf['fuel_tenant'] = None

    conf['fuelclient'] = True  # use fuelclient library by default
    conf['fuel_skip_proxy'] = True
    conf['fuel_logs_remote_dir'] = ['/var/log/docker-logs/remote',
                                    '/var/log/remote']
    conf['fuel_logs_no_remote'] = False  # do not collect /var/log/remote
    '''Do not collect from /var/log/remote/<node>
    if node is in the array of nodes filtered out by soft filter'''
    conf['fuel_logs_exclude_filtered'] = True
    return conf


class Node(BaseNode):
    def get_release(self):
        if self.id == 0:
            cmd = ("awk -F ':' '/release/ {print $2}' "
                   "/etc/nailgun/version.yaml")
        else:
            cmd = ("awk -F ':' '/fuel_version/ {print $2}' "
                   "/etc/astute.yaml")
        release, err, code = tools.ssh_node(ip=self.ip,
                                            command=cmd,
                                            ssh_opts=self.ssh_opts,
                                            timeout=self.timeout,
                                            prefix=self.prefix)
        if code != 0:
            self.logger.warning('%s: could not determine'
                                ' MOS release' % self.repr)
            release = 'n/a'
        else:
            release = release.strip('\n "\'')
        self.logger.info('%s, MOS release: %s' %
                         (self.repr, release))
        return release

    def get_roles_hiera(self):
        def trim_primary(roles):
            trim_roles = [r for r in roles if not r.startswith('primary-')]
            trim_roles += [r[8:] for r in roles if r.startswith('primary-')]
            return trim_roles

        self.logger.debug('%s: roles not defined, trying hiera' % self.repr)
        cmd = 'hiera roles'
        outs, errs, code = tools.ssh_node(ip=self.ip,
                                          command=cmd,
                                          ssh_opts=self.ssh_opts,
                                          env_vars=self.env_vars,
                                          timeout=self.timeout,
                                          prefix=self.prefix)
        self.check_code(code, 'get_roles_hiera', cmd, errs, [0])
        if code == 0:
            try:
                roles = trim_primary(json.loads(outs))
            except:
                self.logger.warning("%s: failed to parse '%s' output as JSON" %
                                    (self.repr, cmd))
                return self.roles
            self.logger.debug('%s: got roles: %s' % (self.repr, roles))
            if roles is not None:
                return roles
            else:
                return self.roles
        else:
            self.logger.warning("%s: failed to load roles via hiera" %
                                self.repr)
            self.roles

    def get_cluster_id(self):
        self.logger.debug('%s: cluster id not defined, trying to determine' %
                          self.repr)
        astute_file = '/etc/astute.yaml'
        cmd = ("python -c 'import yaml; a = yaml.load(open(\"%s\")"
               ".read()); print a[\"cluster\"][\"id\"]'" % astute_file)
        outs, errs, code = tools.ssh_node(ip=self.ip,
                                          command=cmd,
                                          ssh_opts=self.ssh_opts,
                                          env_vars=self.env_vars,
                                          timeout=self.timeout,
                                          prefix=self.prefix)
        return int(outs.rstrip('\n')) if code == 0 else None

    def log_item_manipulate(self, item):
        if self.fuel_logs_no_remote and 'fuel' in self.roles:
            self.logger.debug('adding Fuel remote logs to exclude list')
            if 'exclude' not in item:
                item['exclude'] = []
            for remote_dir in self.fuel_logs_remote_dir:
                item['exclude'].append(remote_dir)
        if 'fuel' in self.roles:
            for n in self.logs_excluded_nodes:
                self.logger.debug('removing remote logs for node:%s' % n)
                if 'exclude' not in item:
                    item['exclude'] = []
                for remote_dir in self.fuel_logs_remote_dir:
                    ipd = os.path.join(remote_dir, n)
                    item['exclude'].append(ipd)


class NodeManager(BaseNodeManager):
    @staticmethod
    def load_conf(filename):
        config = conf.init_default_conf()
        config = add_conf(config)
        config = conf.update_conf(config, filename)
        return config

    def __init__(self, conf, nodes_json=None, logger=None):
        self.base_init(conf, logger)
        self.token = self.conf['fuel_api_token']
        fuelnode = self.fuel_init()
        self.logs_excluded_nodes = []
        if FuelClient and conf['fuelclient']:
            # save os environment variables
            environ = os.environ
            try:
                if self.conf['fuel_skip_proxy']:
                    os.environ['HTTPS_PROXY'] = ''
                    os.environ['HTTP_PROXY'] = ''
                    os.environ['https_proxy'] = ''
                    os.environ['http_proxy'] = ''
                self.logger.info('Setup fuelclient instance')
                self.fuelclient = FuelClient()
                if self.conf['fuel_user']:
                    self.fuelclient.username = self.conf['fuel_user']
                if self.conf['fuel_pass']:
                    self.fuelclient.password = self.conf['fuel_pass']
                if self.conf['fuel_tenant']:
                    self.fuelclient.tenant_name = self.conf['fuel_tenant']
                # self.fuelclient.debug_mode(True)
            except Exception as e:
                self.logger.info('Failed to setup fuelclient instance:%s' % e,
                                 exc_info=True)
                self.fuelclient = None
            os.environ = environ
        else:
            self.logger.info('Skipping setup fuelclient instance')
            self.fuelclient = None
        if nodes_json:
            self.nodes_json = tools.load_json_file(nodes_json)
        else:
            if (not self.get_nodes_fuelclient() and
                    not self.get_nodes_api() and
                    not self.get_nodes_cli()):
                sys.exit(105)
        self.nodes_init(Node)
        # get release information for all nodes
        self.get_release()
        self.post_init()
        fuelnode.logs_excluded_nodes = self.logs_excluded_nodes

    def fuel_init(self):
        if not self.conf['fuel_ip']:
            self.logger.critical('NodeManager: fuel_ip not set')
            sys.exit(106)
        fuelnode = Node(id=0,
                        cluster=0,
                        name='fuel',
                        fqdn='n/a',
                        mac='n/a',
                        os_platform='centos',
                        roles=['fuel'],
                        status='ready',
                        online=True,
                        ip=self.conf['fuel_ip'],
                        conf=self.conf)
        fuelnode.cluster_repr = ""
        fuelnode.repr = "fuel"
        # soft-skip Fuel if it is hard-filtered
        if not self.filter(fuelnode, self.conf['hard_filter']):
            fuelnode.skipped = True
        self.nodes[self.conf['fuel_ip']] = fuelnode
        return fuelnode

    def apply_soft_filter(self):
        # apply soft-filter on all nodes
        for node in self.nodes.values():
            if not self.filter(node, self.conf['soft_filter']):
                node.skipped = True
                if self.conf['fuel_logs_exclude_filtered']:
                    if node.fqdn:
                        self.logs_excluded_nodes.append(node.fqdn)
                    self.logs_excluded_nodes.append(node.ip)

    def get_release(self):
        if (not self.get_release_fuel_client() and
                not self.get_release_api() and
                not self.get_release_cli()):
            self.logger.warning('could not get Fuel and MOS versions')

    def get_nodes_fuelclient(self):
        if not self.fuelclient:
            return False
        try:
            self.logger.info('using fuelclient to get nodes json')
            self.nodes_json = self.fuelclient.get_request('nodes')
            return True
        except Exception as e:
            self.logger.warning(("NodeManager: can't "
                                 "get node list from fuel client:\n%s" % (e)),
                                exc_info=True)
            return False

    def get_release_api(self):
        self.logger.info('getting release via API')
        version_json = self.get_api_request('version')
        if version_json:
            version = json.loads(version_json)
            fuel = self.nodes[self.conf['fuel_ip']]
            fuel.release = version['release']
        else:
            return False
        clusters_json = self.get_api_request('clusters')
        if clusters_json:
            clusters = json.loads(clusters_json)
            self.set_nodes_release(clusters)
            return True
        else:
            return False

    def get_release_fuel_client(self):
        if not self.fuelclient:
            return False
        self.logger.info('getting release via fuelclient')
        try:
            v = self.fuelclient.get_request('version')
            fuel_version = v['release']
            self.logger.debug('version response:%s' % v)
            clusters = self.fuelclient.get_request('clusters')
            self.logger.debug('clusters response:%s' % clusters)
        except:
            self.logger.warning(("Can't get fuel version or "
                                 "clusters information"))
            return False
        self.nodes[self.conf['fuel_ip']].release = fuel_version
        self.set_nodes_release(clusters)
        return True

    def auth_token(self):
        '''Get keystone token to access Nailgun API. Requires Fuel 5+'''
        if self.token:
            return True
        self.logger.info('getting token for Nailgun')
        v2_body = ('{"auth": {"tenantName": "%s", "passwordCredentials": {'
                   '"username": "%s", "password": "%s"}}}')
        # v3 not fully implemented yet
        # v3_body = ('{ "auth": {'
        #            '  "scope": {'
        #            '    "project": {'
        #            '      "name": "%s",'
        #            '      "domain": { "id": "default" }'
        #            '    }'
        #            '  },'
        #            '  "identity": {'
        #            '    "methods": ["password"],'
        #            '    "password": {'
        #            '      "user": {'
        #            '        "name": "%s",'
        #            '        "domain": { "id": "default" },'
        #            '        "password": "%s"'
        #            '      }'
        #            '    }'
        #            '  }'
        #            '}}')
        # Sticking to v2 API for now because Fuel 9.1 has a custom
        # domain_id defined in keystone.conf which we do not know.
        args = {'user': None, 'pass': None, 'tenant': None}
        for a in args:
            if self.conf['fuel_%s' % a]:
                args[a] = self.conf['fuel_%s' % a]
            else:
                args[a] = self.conf['fuel_api_%s' % a]
        req_data = v2_body % (args['tenant'], args['user'], args['pass'])
        req = urllib2.Request("http://%s:%s/v2.0/tokens" %
                              (self.conf['fuel_ip'],
                               self.conf['fuel_api_keystone_port']), req_data,
                              {'Content-Type': 'application/json'})
        try:
            # Disabling v3 token retrieval for now
            # token = urllib2.urlopen(req).info().getheader('X-Subject-Token')
            result = urllib2.urlopen(req)
            resp_body = result.read()
            resp_json = json.loads(resp_body)
            token = resp_json['access']['token']['id']
            self.token = token
            return True
        except:
            return False

    def get_api_request(self, request):
        if self.auth_token():
            url = "http://%s:%s/api/%s" % (self.conf['fuel_ip'],
                                           self.conf['fuel_api_port'],
                                           request)
            req = urllib2.Request(url, None, {'X-Auth-Token': self.token})
            try:
                result = urllib2.urlopen(req)
                code = result.getcode()
                if code == 200:
                    return result.read()
                else:
                    self.logger.error('NodeManager: cannot get API response'
                                      ' from %s, code %s' % (url, code))
            except:
                pass

    def get_nodes_api(self):
        self.logger.info('using API to get nodes json')
        nodes_json = self.get_api_request('nodes')
        if nodes_json:
            self.nodes_json = json.loads(nodes_json)
            return True
        else:
            return False

    def get_nodes_cli(self):
        self.logger.info('using CLI to get nodes json')
        fuelnode = self.nodes[self.conf['fuel_ip']]
        o_auth = n_auth = ''
        entropy = bool(self.conf['fuel_user']) + bool(self.conf['fuel_pass'])
        if entropy == 2:
            # auth for Fuel up to 8.0
            o_auth = '--user %s --password %s' % (self.conf['fuel_user'],
                                                  self.conf['fuel_pass'])
            # Fuel 9.0+
            n_auth = 'OS_USERNAME=%s OS_PASSWORD=%s' % (self.conf['fuel_user'],
                                                        self.conf['fuel_pass'])
        elif entropy == 1:
            self.logger.warning('Must specify both fuel_user and fuel_pass')
        cmd = 'bash -c "%s fuel node --json"' % n_auth
        nodes_json, err, code = tools.ssh_node(ip=fuelnode.ip,
                                               command=cmd,
                                               ssh_opts=fuelnode.ssh_opts,
                                               timeout=fuelnode.timeout,
                                               prefix=fuelnode.prefix)
        if code != 0:
            self.logger.warning(('NodeManager: cannot get fuel node list from'
                                 ' CLI, will fallback. Error: %s') % err)
            cmd = 'bash -c "fuel %s node --json"' % o_auth
            nodes_json, err, code = tools.ssh_node(ip=fuelnode.ip,
                                                   command=cmd,
                                                   ssh_opts=fuelnode.ssh_opts,
                                                   timeout=fuelnode.timeout,
                                                   prefix=fuelnode.prefix)
            if code != 0:
                self.logger.warning(('NodeManager: cannot get '
                                     'fuel node list from CLI: %s') % err)
                self.nodes_json = None
                return False
        self.nodes_json = json.loads(nodes_json)
        return True

    def get_release_cli(self):
        run_items = []
        for key, node in self.selected_nodes.items():
            run_items.append(tools.RunItem(target=node.get_release,
                                           key=key))
        result = tools.run_batch(run_items, 100, dict_result=True)
        if result:
            for key in result:
                self.nodes[key].release = result[key]
            return True
        else:
            return False

    def nodes_init_fallbacks(self):
        self.nodes_get_roles_hiera()
        self.nodes_get_os()
        self.nodes_get_cluster_ids()

    def nodes_get_roles_hiera(self, maxthreads=100):
        run_items = []
        for key, node in self.selected_nodes.items():
            if node.status != 'discover' and not node.roles:
                run_items.append(tools.RunItem(target=node.get_roles_hiera,
                                               key=key))
        result = tools.run_batch(run_items, maxthreads, dict_result=True)
        for key in result:
            if result[key]:
                self.nodes[key].roles = result[key]

    def nodes_get_cluster_ids(self, maxthreads=100):
        self.logger.debug('getting cluster ids from nodes')
        run_items = []
        for key, node in self.selected_nodes.items():
            if not node.cluster:
                run_items.append(tools.RunItem(target=node.get_cluster_id,
                                               key=key))
        result = tools.run_batch(run_items, maxthreads, dict_result=True)
        for key in result:
            if result[key] is not None:
                self.nodes[key].cluster = result[key]

    def set_nodes_release(self, clusters):
        cldict = {}
        for cluster in clusters:
            cldict[cluster['id']] = cluster
        if cldict:
            for node in self.nodes.values():
                if node.cluster:
                    node.release = cldict[node.cluster]['fuel_version']
                else:
                    # set to n/a or may be fuel_version
                    if node.id != 0:
                        node.release = 'n/a'
                self.logger.info('%s: release: %s' % (node.repr, node.release))



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

#    Copyright 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from timmy.nodes import NodeManager as BaseNodeManager


def add_args(parser):
    parser.add_argument('-j', '--nodes-json', required=True,
                        help=('Path to a json file containing host info:'
                              ' ip, roles, etc.'))
    return parser


def check_args(args, conf):
    pass


class NodeManager(BaseNodeManager):
    pass



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

#    Copyright 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
main module
"""

from copy import deepcopy
from datetime import datetime, date, timedelta
from timmy import conf
from timmy.env import project_name, version
from timmy import tools
from tools import w_list, run_with_lock
import logging
import os
import re
import shutil
import sys


class Node(object):
    ckey = 'cmds'
    skey = 'scripts'
    fkey = 'files'
    flkey = 'filelists'
    lkey = 'logs'
    pkey = 'put'
    conf_actionable = [lkey, ckey, skey, fkey, flkey, pkey]
    conf_appendable = [lkey, ckey, skey, fkey, flkey, pkey]
    conf_archive_general = [ckey, skey, fkey, flkey]
    conf_keep_default = [skey, ckey, fkey, flkey]
    conf_once_prefix = 'once_'
    conf_match_prefix = 'by_'
    conf_default_key = '__default'
    header = ['node-id', 'env', 'ip', 'mac', 'os', 'roles', 'online',
              'accessible', 'status', 'name', 'release', 'fqdn']

    def __init__(self, ip, conf, id=None, name=None, fqdn=None, mac=None,
                 cluster=None, roles=None, os_platform=None,
                 online=True, status="ready", logger=None, network_data=None):
        self.logger = logger or logging.getLogger(project_name)
        self.id = int(id) if id is not None else None
        self.mac = mac
        self.cluster = int(cluster) if cluster is not None else None
        if type(roles) is list:
            self.roles = roles
        else:
            self.roles = roles.split(', ') if roles else []
        self.os_platform = os_platform
        self.online = online
        self.status = status
        if ip is None:
            self.logger.critical('Node: ip address must be defined')
            sys.exit(111)
        self.ip = ip
        self.network_data = network_data
        self.release = None
        self.files = []
        self.filelists = []
        self.cmds = []
        self.scripts = []
        # put elements must be tuples - (src, dst)
        self.put = []
        self.data = {}
        self.logsize = 0
        self.mapcmds = {}
        self.mapscr = {}
        self.name = name
        self.fqdn = fqdn
        self.accessible = True
        self.skipped = False
        self.outputs_timestamp = False
        self.outputs_timestamp_dir = None
        self.apply_conf(conf)
        self.cluster_repr = "cluster-%s" % str(cluster)
        if self.id is not None:
            self.repr = "node-%s-%s" % (self.id, self.ip)
        else:
            self.repr = "node-%s" % self.ip

    def __str__(self):
        fields = self.print_table()
        return self.pt.format(*fields)

    def print_table(self):
        if not self.skipped:
            my_id = self.id
        else:
            my_id = '%s [skipped]' % self.id
        return [str(my_id), str(self.cluster), str(self.ip), str(self.mac),
                self.os_platform, ','.join(self.roles),
                str(self.online), str(self.accessible), str(self.status),
                str(self.name), str(self.release), str(self.fqdn)]

    def apply_conf(self, conf, clean=True):

        def apply(k, v, c_a, k_d, o, default=False):
            if k in c_a:
                if any([default,
                        k not in k_d and k not in o,
                        not hasattr(self, k)]):
                    setattr(self, k, deepcopy(w_list(v)))
                else:
                    getattr(self, k).extend(deepcopy(w_list(v)))
                if not default:
                    o[k] = True
            else:
                setattr(self, k, deepcopy(v))

        def r_apply(el, p, c_a, k_d, o, d, clean=False):
            # apply normal attributes
            for k in [k for k in el if not k.startswith(p)]:
                if el == conf and clean:
                    apply(k, el[k], c_a, k_d, o, default=True)
                else:
                    apply(k, el[k], c_a, k_d, o)
            # apply match attributes
            for k in [k for k in el if k.startswith(p)]:
                attr_name = k[len(p):]
                if hasattr(self, attr_name):
                    attr = w_list(getattr(self, attr_name))
                    matching_keys = []
                    # negative matching ("no_")
                    for nk in [nk for nk in el[k] if nk.startswith('no_')]:
                        key = nk[4:]
                        if key not in attr:
                            matching_keys.append(nk)
                    # positive matching
                    for v in attr:
                        if v in el[k]:
                            matching_keys.append(v)
                    # apply matching keys
                    for mk in matching_keys:
                        subconf = el[k][mk]
                        if d in el:
                            d_conf = el[d]
                            for a in d_conf:
                                apply(a, d_conf[a], c_a, k_d, o)
                        r_apply(subconf, p, c_a, k_d, o, d)

        p = Node.conf_match_prefix
        c_a = Node.conf_appendable
        k_d = Node.conf_keep_default
        d = Node.conf_default_key
        overridden = {}
        if clean:
            '''clean appendable keep_default params to ensure no content
            duplication if this function gets called more than once'''
            for f in set(c_a).intersection(k_d):
                setattr(self, f, [])
        r_apply(conf, p, c_a, k_d, overridden, d, clean=clean)

    def get_os(self):
        self.logger.debug('%s: os_platform not defined, trying to determine' %
                          self.repr)
        cmd = 'which lsb_release'
        outs, errs, code = tools.ssh_node(ip=self.ip,
                                          command=cmd,
                                          ssh_opts=self.ssh_opts,
                                          env_vars=self.env_vars,
                                          timeout=self.timeout,
                                          prefix=self.prefix)
        return 'centos' if code else 'ubuntu'

    def check_access(self):
        self.logger.debug('%s: verifyng node access' %
                          self.repr)
        cmd = 'true'
        outs, errs, code = tools.ssh_node(ip=self.ip,
                                          command=cmd,
                                          ssh_opts=self.ssh_opts,
                                          env_vars=self.env_vars,
                                          timeout=self.timeout,
                                          prefix=self.prefix)
        if code == 0:
            return True
        else:
            self.logger.info('%s: not accessible' % self.repr)
            return False

    def exec_cmd(self, fake=False, ok_codes=None):
        cl = self.cluster_repr
        self.logger.debug('%s/%s/%s/%s' %
                          (self.outdir, Node.ckey, cl, self.repr))
        if self.cmds:
            ddir = os.path.join(self.outdir, Node.ckey, cl, self.repr)
            tools.mdir(ddir)
            self.cmds = sorted(self.cmds)
        mapcmds = {}
        for c in self.cmds:
            for cmd in c:
                dfile = os.path.join(ddir, cmd)
                if self.outputs_timestamp:
                        dfile += self.outputs_timestamp_str
                self.logger.info('outfile: %s' % dfile)
                mapcmds[cmd] = dfile
                if not fake:
                    bash_cmd = "bash -c '%s'" % c[cmd]
                    outs, errs, code = tools.ssh_node(ip=self.ip,
                                                      command=bash_cmd,
                                                      ssh_opts=self.ssh_opts,
                                                      env_vars=self.env_vars,
                                                      timeout=self.timeout,
                                                      prefix=self.prefix)
                    self.check_code(code, 'exec_cmd', c[cmd], errs, ok_codes)
                    try:
                        with open(dfile, 'w') as df:
                            df.write(outs.encode('utf-8'))
                    except:
                        self.logger.error("can't write to file %s" %
                                          dfile)
        if self.scripts:
            ddir = os.path.join(self.outdir, Node.skey, cl, self.repr)
            tools.mdir(ddir)
            self.scripts = sorted(self.scripts)
        mapscr = {}
        for scr in self.scripts:
            if type(scr) is dict:
                env_vars = scr.values()[0]
                scr = scr.keys()[0]
            else:
                env_vars = self.env_vars
            if os.path.sep in scr:
                f = scr
            else:
                f = os.path.join(self.rqdir, Node.skey, scr)
            self.logger.debug('%s, exec: %s' % (self.repr, f))
            dfile = os.path.join(ddir, os.path.basename(f))
            if self.outputs_timestamp:
                    dfile += self.outputs_timestamp_str
            self.logger.debug('outfile: %s' % dfile)
            mapscr[scr] = dfile
            if not fake:
                outs, errs, code = tools.ssh_node(ip=self.ip,
                                                  filename=f,
                                                  ssh_opts=self.ssh_opts,
                                                  env_vars=env_vars,
                                                  timeout=self.timeout,
                                                  prefix=self.prefix)
                self.check_code(code, 'exec_cmd', 'script %s' % f, errs,
                                ok_codes)
                try:
                    with open(dfile, 'w') as df:
                        df.write(outs.encode('utf-8'))
                except:
                    self.logger.error("can't write to file %s" % dfile)
        return mapcmds, mapscr

    def exec_simple_cmd(self, cmd, timeout=15, infile=None, outfile=None,
                        fake=False, ok_codes=None, input=None, decode=True):
        self.logger.info('%s, exec: %s' % (self.repr, cmd))
        if not fake:
            outs, errs, code = tools.ssh_node(ip=self.ip,
                                              command=cmd,
                                              ssh_opts=self.ssh_opts,
                                              env_vars=self.env_vars,
                                              timeout=timeout,
                                              outputfile=outfile,
                                              ok_codes=ok_codes,
                                              decode=decode,
                                              input=input,
                                              prefix=self.prefix)
            self.check_code(code, 'exec_simple_cmd', cmd, errs, ok_codes)

    def exec_pair(self, phase, server_node=None, fake=False):
        sn = server_node
        cl = self.cluster_repr
        if sn:
            self.logger.debug('%s: phase %s: server %s' % (self.repr, phase,
                                                           sn.repr))
        else:
            self.logger.debug('%s: phase %s' % (self.repr, phase))
        nond_msg = ('%s: network specified but network_data not set for %s')
        nonet_msg = ('%s: network %s not found in network_data of %s')
        nosrv_msg = ('%s: server_node not provided')
        noip_msg = ('%s: %s has no IP in network %s')
        for i in self.scripts_all_pairs:
            if phase not in i:
                self.logger.warning('phase %s not defined in config' % phase)
                return self.scripts_all_pairs
            if phase.startswith('client'):
                if not sn:
                    self.logger.warning(nosrv_msg % self.repr)
                    return self.scripts_all_pairs
                if 'network' in i:
                    if not sn.network_data:
                        self.logger.warning(nond_msg % (self.repr, sn.repr))
                        return self.scripts_all_pairs
                    nd = sn.network_data
                    net_dict = dict((v['name'], v) for v in nd)
                    if i['network'] not in net_dict:
                        self.logger.warning(nonet_msg % (self.repr, sn.repr))
                        return self.scripts_all_pairs
                    if 'ip' not in net_dict[i['network']]:
                        self.logger.warning(noip_msg % (self.repr, sn.repr,
                                                        i['network']))
                        return self.scripts_all_pairs
                    ip = net_dict[i['network']]['ip']
                    if '/' in ip:
                        server_ip = ip.split('/')[0]
                    else:
                        server_ip = ip
                else:
                    server_ip = sn.ip
            phase_val = i[phase]
            ddir = os.path.join(self.outdir, 'scripts_all_pairs', cl, phase,
                                self.repr)
            tools.mdir(ddir)
            if type(phase_val) is dict:
                env_vars = [phase_val.values()[0]]
                phase_val = phase_val.keys()[0]
            else:
                env_vars = self.env_vars
            if os.path.sep in phase_val:
                f = phase_val
            else:
                f = os.path.join(self.rqdir, Node.skey, phase_val)
            dfile = os.path.join(ddir, os.path.basename(f))
            if phase.startswith('client'):
                env_vars.append('SERVER_IP=%s' % server_ip)
                dname = os.path.basename(f) + '-%s' % server_ip
                dfile = os.path.join(ddir, dname)
            elif phase == 'server_stop' and 'server_output' in i:
                env_vars.append('SERVER_OUTPUT=%s' % i['server_output'])
            if fake:
                return self.scripts_all_pairs
            outs, errs, code = tools.ssh_node(ip=self.ip,
                                              filename=f,
                                              ssh_opts=self.ssh_opts,
                                              env_vars=env_vars,
                                              timeout=self.timeout,
                                              prefix=self.prefix)
            self.check_code(code, 'exec_pair, phase:%s' % phase, f, errs)
            if phase == 'server_start' and code == 0:
                i['server_output'] = outs.strip()
            open(dfile, 'a+').write(outs)
        return self.scripts_all_pairs

    def get_files(self, timeout=15):
        self.logger.info('%s: getting files' % self.repr)
        cl = self.cluster_repr
        if self.files or self.filelists:
            ddir = os.path.join(self.outdir, Node.fkey, cl, self.repr)
            tools.mdir(ddir)
        data = ''
        for f in self.filelists:
            if os.path.sep in f:
                fname = f
            else:
                fname = os.path.join(self.rqdir, Node.flkey, f)
            try:
                with open(fname, 'r') as df:
                    for line in df:
                        if not line.isspace() and line[0] != '#':
                            data += line
            except:
                self.logger.error('could not read file: %s' % fname)
        self.logger.debug('%s: data:\n%s' % (self.repr, data))
        if data:
            o, e, c = tools.get_files_rsync(ip=self.ip,
                                            data=data,
                                            ssh_opts=self.ssh_opts,
                                            rsync_opts=self.rsync_opts,
                                            dpath=ddir,
                                            timeout=self.timeout)
            self.check_code(c, 'get_files', 'tools.get_files_rsync', e)
        for f in self.files:
            outs, errs, code = tools.get_file_scp(ip=self.ip,
                                                  file=f,
                                                  ssh_opts=self.ssh_opts,
                                                  ddir=ddir,
                                                  recursive=True)
            self.check_code(code, 'get_files', 'tools.get_file_scp', errs)

    def put_files(self):
        self.logger.info('%s: putting files' % self.repr)
        for f in self.put:
            outs, errs, code = tools.put_file_scp(ip=self.ip,
                                                  file=f[0],
                                                  dest=f[1],
                                                  ssh_opts=self.ssh_opts,
                                                  recursive=True)
            self.check_code(code, 'put_files', 'tools.put_file_scp', errs)

    def log_item_manipulate(self, item):
        pass

    def logs_populate(self, timeout=5):

        def filter_by_re(item, string):
            return (('include' not in item or not item['include'] or
                     any([re.search(i, string) for i in item['include']])) and
                    ('exclude' not in item or not item['exclude'] or not
                     any([re.search(e, string) for e in item['exclude']])))

        for item in self.logs:
            self.log_item_manipulate(item)
            start_str = None
            if 'start' in item or hasattr(self, 'logs_days'):
                if hasattr(self, 'logs_days') and 'start' not in item:
                    start = self.logs_days
                else:
                    start = item['start']
                if any([type(start) is str and re.match(r'-?\d+', start),
                        type(start) is int]):
                    days = abs(int(str(start)))
                    start_str = str(date.today() - timedelta(days=days))
                else:
                    for format in ['%Y-%m-%d', '%Y-%m-%d %H:%M:%S']:
                        try:
                            if datetime.strptime(start, format):
                                start_str = start
                                break
                        except ValueError:
                            pass
                    if not start_str:
                        self.logger.warning(('incorrect value of "start"'
                                             ' parameter in "logs": "%s" -'
                                             ' ignoring...')
                                            % start)
            if start_str:
                start_param = ' -newermt "$(date -d \'%s\')"' % start_str
            else:
                start_param = ''
            cmd = ("find '%s' -type f%s -exec du -b {} +" % (item['path'],
                                                             start_param))
            self.logger.info('%s: logs du-cmd: %s' %
                             (self.repr, cmd))
            outs, errs, code = tools.ssh_node(ip=self.ip,
                                              command=cmd,
                                              ssh_opts=self.ssh_opts,
                                              env_vars=self.env_vars,
                                              timeout=timeout,
                                              prefix=self.prefix)
            if code == 124:
                self.logger.error("%s: command: %s, "
                                  "timeout code: %s, error message: %s" %
                                  (self.repr, cmd, code, errs))
                break
            if len(outs):
                item['files'] = {}
                for line in outs.split('\n'):
                    if '\t' in line:
                        size, f = line.split('\t')
                        if filter_by_re(item, f):
                            item['files'][f] = int(size)
                        else:
                            self.logger.debug('log file "%s" excluded' % f)
                self.logger.debug('logs: %s' % (item['files']))
            self.logger.info('%s: total logs size: %dMB' %
                             (self.repr,
                              sum(self.logs_dict().values())/1024/1024))
        return self.logs

    def logs_dict(self):
        result = {}
        for item in self.logs:
            if 'files' in item:
                for f, s in item['files'].items():
                    if f in result:
                        result[f] = max(result[f], s)
                    else:
                        result[f] = s
        return result

    def check_code(self, code, func_name, cmd, err, ok_codes=None):
        if code:
            if not ok_codes or code not in ok_codes:
                self.logger.warning("%s: func: %s: "
                                    "cmd: '%s' exited %d, error: %s" %
                                    (self.repr, func_name, cmd, code, err))

    def get_results(self, result_map):
        # result_map should be either mapcmds or mapscr
        if self.id is not None:
            short_repr = "node-%s" % self.id
        else:
            short_repr = self.ip
        output = []
        for cmd in sorted(result_map):
            with open(result_map[cmd], 'r') as f:
                for line in f.readlines():
                    output.append(line.rstrip('\n'))
        return short_repr, output


class NodeManager(object):
    """Class NodeManager """

    @staticmethod
    def load_conf(filename):
        config = conf.init_default_conf()
        config = conf.update_conf(config, filename)
        return config

    def __init__(self, conf, nodes_json, logger=None):
        self.base_init(conf, logger)
        self.nodes_json = tools.load_json_file(nodes_json)
        self.nodes_init(Node)
        self.post_init()

    def nodes_init_fallbacks(self):
        self.nodes_get_os()

    def base_init(self, conf, logger=None):
        self.conf = conf
        self.logger = logger or logging.getLogger(project_name)
        if conf['outputs_timestamp'] or conf['dir_timestamp']:
            timestamp_str = datetime.now().strftime('_%F_%H-%M-%S')
            if conf['outputs_timestamp']:
                conf['outputs_timestamp_str'] = timestamp_str
            if conf['dir_timestamp']:
                conf['outdir'] += timestamp_str
                conf['archive_dir'] += timestamp_str
        if conf['clean']:
            shutil.rmtree(conf['outdir'], ignore_errors=True)
            tools.mdir(conf['outdir'])
        version_filename = '%s_version.txt' % project_name
        version_filepath = os.path.join(conf['outdir'], version_filename)
        with open(version_filepath, 'a') as f:
            now = datetime.now()
            ver_msg = 'running timmy version %s' % version
            f.write('%s: %s\n' % (now, ver_msg))
            self.logger.info(ver_msg)
        if not conf['shell_mode']:
            self.rqdir = conf['rqdir']
            if (not os.path.exists(self.rqdir)):
                self.logger.critical(('NodeManager: directory %s does not'
                                      ' exist') % self.rqdir)
                sys.exit(101)
            if self.conf['rqfile']:
                self.import_rq()
        self.nodes = {}

    def apply_soft_filter(self):
        # apply soft-filter on all nodes
        for node in self.nodes.values():
            if not self.filter(node, self.conf['soft_filter']):
                node.skipped = True

    def post_init(self):
        self.nodes_reapply_conf()
        self.apply_soft_filter()
        self.conf_assign_once()

    def __str__(self):
        def ml_column(matrix, i):
            a = [row[i] for row in matrix]
            mc = 0
            for word in a:
                lw = len(str(word))
                mc = lw if (lw > mc) else mc
            return mc + 2
        header = Node.header
        nodestrings = [header]
        for n in self.sorted_nodes():
            nodestrings.append(n.print_table())
        colwidth = []
        for i in range(len(header)):
            colwidth.append(ml_column(nodestrings, i))
        pt = ''
        for i in range(len(colwidth)):
            pt += '{%s:<%s}' % (i, str(colwidth[i]))
        nodestrings = [(pt.format(*header))]
        for n in self.sorted_nodes():
            n.pt = pt
            nodestrings.append(str(n))
        return '\n'.join(nodestrings)

    def sorted_nodes(self):
        nv = self.nodes.values()
        s = [n for n in sorted(nv, key=lambda x: (x.id, x.ip))]
        return s

    def import_rq(self):

        def sub_is_match(el, d, p, once_p):
            if type(el) is not dict:
                return False
            checks = []
            for i in el:
                checks.append(any([i == d,
                                  i.startswith(p),
                                  i.startswith(once_p)]))
            return all(checks)

        def r_sub(attr, el, k, d, p, once_p, dst):
            match_sect = False
            if type(k) is str and (k.startswith(p) or k.startswith(once_p)):
                match_sect = True
            if k not in dst and k != attr:
                dst[k] = {}
            if d in el[k]:
                if k == attr:
                    if k in Node.conf_appendable:
                        dst[k] = w_list(dst[k])
                        dst[k] += w_list(el[k][d])
                    else:
                        dst[k] = el[k][d]
                elif k.startswith(p) or k.startswith(once_p):
                    dst[k][d] = {attr: el[k][d]}
                else:
                    dst[k][attr] = el[k][d]
            if k == attr:
                subks = [subk for subk in el[k] if subk != d]
                for subk in subks:
                    r_sub(attr, el[k], subk, d, p, once_p, dst)
            elif match_sect or sub_is_match(el[k], d, p, once_p):
                subks = [subk for subk in el[k] if subk != d]
                for subk in subks:
                    if el[k][subk] is not None:
                        if subk not in dst[k]:
                            dst[k][subk] = {}
                        r_sub(attr, el[k], subk, d, p, once_p, dst[k])
            else:
                dst[k][attr] = el[k]

        def merge_rq(rqfile, dst):
            file = rqfile['file']
            if os.path.sep in file:
                src = tools.load_yaml_file(file)
            else:
                f = os.path.join(self.rqdir, file)
                src = tools.load_yaml_file(f)
            if self.conf['logs_no_default'] and rqfile['default']:
                if 'logs' in src:
                    src.pop('logs')
            p = Node.conf_match_prefix
            once_p = Node.conf_once_prefix + p
            d = Node.conf_default_key
            for attr in src:
                r_sub(attr, src, attr, d, p, once_p, dst)

        dst = self.conf
        for rqfile in self.conf['rqfile']:
            merge_rq(rqfile, dst)

    def nodes_init(self, NodeClass):
        for node_data in self.nodes_json:
            params = {'conf': self.conf}
            keys = ['id', 'cluster', 'roles', 'fqdn', 'name', 'mac',
                    'os_platform', 'status', 'online', 'ip', 'network_data']
            for key in keys:
                if key in node_data:
                    params[key] = node_data[key]
            node = NodeClass(**params)
            if self.filter(node, self.conf['hard_filter']):
                self.nodes[node.ip] = node
        self.nodes_check_access()
        self.nodes_init_fallbacks()

    def conf_assign_once(self):
        once = Node.conf_once_prefix
        p = Node.conf_match_prefix
        once_p = once + p
        for k in [k for k in self.conf if k.startswith(once)]:
            attr_name = k[len(once_p):]
            assigned = dict((k, None) for k in self.conf[k])
            for ak in assigned:
                for node in self.selected_nodes.values():
                    if hasattr(node, attr_name) and not assigned[ak]:
                        attr = w_list(getattr(node, attr_name))
                        for v in attr:
                            if v == ak:
                                once_conf = self.conf[k][ak]
                                node.apply_conf(once_conf, clean=False)
                                assigned[ak] = node.ip
                                break
                    if assigned[ak]:
                        break

    def nodes_reapply_conf(self):
        for node in self.nodes.values():
            node.apply_conf(self.conf)

    def nodes_get_os(self, maxthreads=100):
        run_items = []
        for key, node in self.selected_nodes.items():
            if not node.os_platform:
                run_items.append(tools.RunItem(target=node.get_os, key=key))
        result = tools.run_batch(run_items, maxthreads, dict_result=True)
        for key in result:
            if result[key]:
                self.nodes[key].os_platform = result[key]

    def nodes_check_access(self, maxthreads=100):
        self.logger.debug('checking if nodes are accessible')
        run_items = []
        for key, node in self.selected_nodes.items():
            run_items.append(tools.RunItem(target=node.check_access,
                                           key=key))
        result = tools.run_batch(run_items, maxthreads, dict_result=True)
        for key in result:
            self.nodes[key].accessible = result[key]

    def filter(self, node, node_filter):
        f = node_filter
        if f is self.conf['soft_filter'] and not node.accessible:
            return False
        # soft-skip Fuel node for shell mode
        if (node.id == 0 and self.conf['shell_mode']):
            return False
        else:
            elems = []
            for k in f:
                if k.startswith('no_') and hasattr(node, k[3:]):
                    elems.append({'node_k': k[3:], 'k': k, 'negative': True})
                elif hasattr(node, k) and f[k]:
                    elems.append({'node_k': k, 'k': k, 'negative': False})
            checks = []
            for el in elems:
                node_v = w_list(getattr(node, el['node_k']))
                filter_v = w_list(f[el['k']])
                if el['negative']:
                    checks.append(set(node_v).isdisjoint(filter_v))
                elif node.id != 0:
                    '''Do not apply normal (positive) filters to Fuel node
                    , Fuel node will only be filtered by negative filters
                    such as no_id = [0] or no_roles = ['fuel']'''
                    checks.append(not set(node_v).isdisjoint(filter_v))
            return all(checks)

    @run_with_lock
    def run_commands(self, timeout=15, fake=False, maxthreads=100):
        run_items = []
        for key, node in self.selected_nodes.items():
            run_items.append(tools.RunItem(target=node.exec_cmd,
                                           args={'fake': fake},
                                           key=key))
        result = tools.run_batch(run_items, maxthreads, dict_result=True)
        for key in result:
            self.nodes[key].mapcmds = result[key][0]
            self.nodes[key].mapscr = result[key][1]

    def calculate_log_size(self, timeout=15, maxthreads=100):
        total_size = 0
        run_items = []
        for key, node in self.selected_nodes.items():
            run_items.append(tools.RunItem(target=node.logs_populate,
                                           args={'timeout': timeout},
                                           key=key))
        result = tools.run_batch(run_items, maxthreads, dict_result=True)
        for key in result:
            self.nodes[key].logs = result[key]
        for node in self.selected_nodes.values():
            total_size += sum(node.logs_dict().values())
        self.logger.info('Full log size on nodes(with fuel): %d bytes' %
                         total_size)
        self.alogsize = total_size
        return self.alogsize

    def is_enough_space(self):
        tools.mdir(self.conf['archive_dir'])
        outs, errs, code = tools.free_space(self.conf['archive_dir'],
                                            timeout=1)
        if code != 0:
            self.logger.error("Can't get free space: %s" % errs)
            return False
        try:
            fs = int(outs.rstrip('\n'))
        except:
            self.logger.error("can't get free space\nouts: %s" %
                              outs)
            return False
        coeff = self.conf['logs_size_coefficient']
        self.logger.info('logsize: %dMB * %s, free space: %dMB' %
                         (self.alogsize/1024/1024, coeff, fs/1024))
        if (self.alogsize*coeff > fs*1024):
            self.logger.error('Not enough space in "%s", logsize: %dMB * %s, '
                              'available: %dMB. Decrease logs_size_coefficient'
                              ' config parameter (--logs-coeff CLI parameter)'
                              ' or free up space.' % (self.conf['archive_dir'],
                                                      self.alogsize/1024/1024,
                                                      coeff,
                                                      fs/1024))
            return False
        else:
            return True

    @run_with_lock
    def create_archive_general(self, timeout):
        if not os.path.isdir(self.conf['outdir']):
            self.logger.warning("Nothing to do, directory %s doesn't exist" %
                                self.conf['outdir'])
            return
        outfile = os.path.join(self.conf['archive_dir'],
                               self.conf['archive_name'])
        cmd = "tar zcf '%s' -C %s %s" % (outfile, self.conf['outdir'], ".")
        tools.mdir(self.conf['archive_dir'])
        self.logger.debug("cmd: %s" % cmd)
        outs, errs, code = tools.launch_cmd(cmd, timeout)
        if code != 0:
            self.logger.error("Can't create archive %s" % (errs))

    def find_adm_interface_speed(self):
        '''Returns interface speed through which logs will be dowloaded'''
        for node in self.selected_nodes.values():
            if not (node.ip == 'localhost' or node.ip.startswith('127.')):
                cmd = ("%s$(/sbin/ip -o route get %s | cut -d' ' -f3)/speed" %
                       ('cat /sys/class/net/', node.ip))
                out, err, code = tools.launch_cmd(cmd, node.timeout)
                if code != 0:
                    self.logger.warning("can't get iface speed: err: %s" % err)
                    return self.conf['logs_speed_default']
                try:
                    speed = int(out)
                except:
                    speed = self.conf['logs_speed_default']
                return speed

    @run_with_lock
    def get_logs(self, timeout, fake=False, maxthreads=10):
        if fake:
            self.logger.info('fake = True, skipping')
            return
        if self.conf['logs_speed_limit']:
            if self.conf['logs_speed'] > 0:
                speed = self.conf['logs_speed']
            else:
                speed = self.find_adm_interface_speed()
            speed = int(speed * 0.9 / min(maxthreads, len(self.nodes)))
            py_slowpipe = tools.slowpipe % speed
            limitcmd = "| python -c '%s'; exit ${PIPESTATUS}" % py_slowpipe
        run_items = []
        for node in self.selected_nodes.values():
            if not node.logs_dict():
                self.logger.info(("%s: no logs to collect") % node.repr)
                continue
            node.archivelogsfile = os.path.join(self.conf['archive_dir'],
                                                'logs-%s.tar.gz' % node.repr)
            tools.mdir(self.conf['archive_dir'])
            input = ''
            for fn in node.logs_dict():
                input += '%s\0' % fn.lstrip(os.path.abspath(os.sep))
            cmd = ("tar --transform 's,^,%s/,' --gzip -C %s --create "
                   "--warning=no-file-changed --file - --null --files-from -" %
                   (node.repr, os.path.abspath(os.sep)))
            if self.conf['logs_speed_limit']:
                if not (node.ip == 'localhost' or node.ip.startswith('127.')):
                    cmd = ' '.join([cmd, limitcmd])
            args = {'cmd': cmd,
                    'timeout': timeout,
                    'outfile': node.archivelogsfile,
                    'input': input,
                    'ok_codes': [0, 1],
                    'decode': False}
            run_items.append(tools.RunItem(target=node.exec_simple_cmd,
                                           args=args))
        tools.run_batch(run_items, maxthreads)

    @run_with_lock
    def get_files(self, timeout=15):
        run_items = []
        for node in self.selected_nodes.values():
            run_items.append(tools.RunItem(target=node.get_files))
        tools.run_batch(run_items, 10)

    @run_with_lock
    def put_files(self):
        run_items = []
        for node in self.selected_nodes.values():
            run_items.append(tools.RunItem(target=node.put_files))
        tools.run_batch(run_items, 10)

    @run_with_lock
    def run_scripts_all_pairs(self, maxthreads, fake=False):
        if len(self.selected_nodes) < 2:
            self.logger.warning('less than 2 nodes are available, '
                                'skipping paired scripts')
            return
        run_server_start_items = []
        run_server_stop_items = []
        for n in self.selected_nodes.values():
            start_args = {'phase': 'server_start', 'fake': fake}
            run_server_start_items.append(tools.RunItem(target=n.exec_pair,
                                                        args=start_args,
                                                        key=n.ip))
            stop_args = {'phase': 'server_stop', 'fake': fake}
            run_server_stop_items.append(tools.RunItem(target=n.exec_pair,
                                                       args=stop_args))
        result = tools.run_batch(run_server_start_items, maxthreads,
                                 dict_result=True)
        for key in result:
            self.nodes[key].scripts_all_pairs = result[key]
        for pairset in tools.all_pairs(self.selected_nodes.values()):
            run_client_items = []
            self.logger.info(['%s->%s' % (p[0].ip, p[1].ip) for p in pairset])
            for pair in pairset:
                client = pair[0]
                server = pair[1]
                client_args = {'phase': 'client', 'server_node': server,
                               'fake': fake}
                run_client_items.append(tools.RunItem(target=client.exec_pair,
                                                      args=client_args))
            tools.run_batch(run_client_items, len(run_client_items))
        tools.run_batch(run_server_stop_items, maxthreads)

    def has(self, *keys):
        nodes = {}
        for k in keys:
            for n in self.nodes.values():
                if hasattr(n, k):
                    attr = getattr(n, k)
                    if attr:
                        if k not in nodes:
                            nodes[k] = []
                        nodes[k].append(n)
        return nodes

    @property
    def selected_nodes(self):
        return dict([(ip, n) for ip, n in self.nodes.items() if not n.skipped])


def main(argv=None):
    return 0

if __name__ == '__main__':
    sys.exit(main(sys.argv))



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

#    Copyright 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from timmy.env import project_name, version
from timmy.nodes import Node
from timmy.tools import signal_wrapper
import argparse
import logging
import logging.handlers
import os
import sys


def pretty_run(quiet, msg, f, args=[], kwargs={}):
    if not quiet:
        sys.stdout.write('%s...\r' % msg)
        sys.stdout.flush()
    result = f(*args, **kwargs)
    if not quiet:
        print('%s: done' % msg)
    return result


def add_args(parser, module):
    parser = module.add_args(parser)
    return parser


def parser_init(add_help=False):
    desc = 'Parallel remote command execution and file manipulation tool'
    parser = argparse.ArgumentParser(description=desc, add_help=add_help)
    parser.add_argument('-V', '--version', action='store_true',
                        help='Print Timmy version and exit.')
    parser.add_argument('-c', '--config',
                        help='Path to a YAML configuration file.')
    parser.add_argument('-o', '--dest-file',
                        help=('Output filename for the archive in tar.gz'
                              ' format for command outputs and collected'
                              ' files. Overrides "archive_" config options.'
                              ' If logs are collected they will be placed'
                              ' in the same folder (but separate archives).'))
    parser.add_argument('--log-file', default=None,
                        help='Redirect Timmy log to a file.')
    parser.add_argument('-e', '--env', type=int,
                        help='Env ID. Run only on specific environment.')
    parser.add_argument('-r', '--role', action='append',
                        help=('Can be specified multiple times.'
                              ' Run only on the specified role.'))
    parser.add_argument('-i', '--id', type=int, action='append',
                        help=('Can be specified multiple times.'
                              ' Run only on the node(s) with given IDs.'))
    parser.add_argument('-d', '--days', type=int, metavar='NUMBER',
                        help=('Define log collection period in days.'
                              ' Timmy will collect only logs updated on or'
                              ' more recently then today minus the given'
                              ' number of days. Default - 30.'))
    parser.add_argument('-G', '--get', action='append', metavar='PATH',
                        help=('Enables shell mode. Can be specified multiple'
                              ' times. Filemask to collect via "scp -r".'
                              ' Result is placed into a folder specified'
                              ' by "outdir" config option.'
                              ' For help on shell mode, read timmy/conf.py.'))
    parser.add_argument('-C', '--command', action='append',
                        help=('Enables shell mode. Can be specified'
                              ' multiple times. Shell command to execute.'
                              ' For help on shell mode, read timmy/conf.py.'))
    parser.add_argument('-S', '--script', action='append',
                        help=('Enables shell mode. Can be specified'
                              ' multiple times. Bash script name to execute.'
                              ' Script must be placed in "%s" folder inside'
                              ' a path specified by "rqdir" configuration'
                              ' parameter. For help on shell mode, read'
                              ' timmy/conf.py.') % Node.skey)
    parser.add_argument('-P', '--put', nargs=2, action='append',
                        metavar=('SOURCE', 'DESTINATION'),
                        help=('Enables shell mode. Can be specified multiple'
                              ' times. Upload filemask via"scp -r" to node(s).'
                              ' Each argument must contain two strings -'
                              ' source file/path/mask and dest. file/path.'
                              ' For help on shell mode, read timmy/conf.py.'))
    parser.add_argument('-L', '--get-logs', nargs=3, action='append',
                        metavar=('PATH', 'INCLUDE', 'EXCLUDE'),
                        help=('Define specific logs to collect. Implies "-l".'
                              ' Each -L option requires 3 values in the'
                              ' following order: path, include, exclude.'
                              ' See configuration doc for details on each of'
                              ' these parameters. Values except path can be'
                              ' skipped by passing empty strings. Example: -L'
                              ' "/var/mylogs/" "" "exclude-string"'))
    parser.add_argument('--rqfile', metavar='PATH', action='append',
                        help=('Can be specified multiple times. Path to'
                              ' rqfile(s) in yaml format, overrides default.'))
    parser.add_argument('-l', '--logs', action='store_true',
                        help=('Collect logs from nodes. Logs are not collected'
                              ' by default due to their size.'))
    parser.add_argument('--logs-no-default', action='store_true',
                        help=('Do not use default log collection parameters,'
                              ' only use what has been provided either via -L'
                              ' or in rqfile(s). Implies "-l".'))
    parser.add_argument('--logs-speed', type=int, metavar='MBIT/S',
                        help=('Limit log collection bandwidth to 90%% of the'
                              ' specified speed in Mbit/s.'))
    parser.add_argument('--logs-speed-auto', action='store_true',
                        help=('Limit log collection bandwidth to 90%% of local'
                              ' admin interface speed. If speed detection'
                              ' fails, a default value will be used. See'
                              ' "logs_speed_default" in conf.py.'))
    parser.add_argument('--logs-coeff', type=float, metavar='RATIO',
                        help=('Estimated logs compression ratio - this value'
                              ' is used during free space check. Set to a'
                              ' lower value (default - 1.05) to collect logs'
                              ' of a total size larger than locally available'
                              '. Values lower than 0.3 are not recommended'
                              ' and may result in filling up local disk.'))
    parser.add_argument('--only-logs',
                        action='store_true',
                        help=('Only collect logs, do not run commands or'
                              ' collect files.'))
    parser.add_argument('--fake-logs',
                        help='Do not collect logs, only calculate size.',
                        action='store_true')
    parser.add_argument('--no-archive',
                        help=('Do not create results archive. By default,'
                              ' an archive with all outputs and files'
                              ' is created every time you run Timmy.'),
                        action='store_true')
    parser.add_argument('--no-clean',
                        help=('Do not clean previous results. Allows'
                              ' accumulating results across runs.'),
                        action='store_true')
    parser.add_argument('-q', '--quiet',
                        help=('Print only command execution results and log'
                              ' messages. Good for quick runs / "watch" wrap.'
                              ' This option disables any -v parameters.'),
                        action='store_true')
    parser.add_argument('--maxthreads', type=int, default=100,
                        metavar='NUMBER',
                        help=('Maximum simultaneous nodes for command'
                              'execution.'))
    parser.add_argument('--logs-maxthreads', type=int, default=10,
                        metavar='NUMBER',
                        help='Maximum simultaneous nodes for log collection.')
    parser.add_argument('-t', '--outputs-timestamp',
                        help=('Add timestamp to outputs - allows accumulating'
                              ' outputs of identical commands/scripts across'
                              ' runs. Only makes sense with --no-clean for'
                              ' subsequent runs.'),
                        action='store_true')
    parser.add_argument('-T', '--dir-timestamp',
                        help=('Add timestamp to output folders (defined by'
                              ' "outdir" and "archive_dir" config options).'
                              ' Makes each run store results in new folders.'
                              ' This way Timmy will always preserve previous'
                              ' results. Do not forget to clean up the results'
                              ' manually when using this option.'),
                        action='store_true')
    parser.add_argument('-m', '--module', metavar='INVENTORY MODULE',
                        default='fuel',
                        help='Use module to get node data')
    parser.add_argument('-v', '--verbose', action='count', default=0,
                        help=('This works for -vvvv, -vvv, -vv, -v, -v -v,'
                              'etc, If no -v then logging.WARNING is '
                              'selected if more -v are provided it will '
                              'step to INFO and DEBUG unless the option '
                              '-q(--quiet) is specified'))
    return parser


@signal_wrapper
def main(argv=None):
    if argv is None:
        argv = sys.argv
    parser = parser_init()
    args, unknown = parser.parse_known_args(argv[1:])
    parser = parser_init(add_help=True)
    if args.module:
        inventory = __import__('timmy.modules.%s' % args.module,
                               fromlist=['timmy.modules'])
        parser = add_args(parser, inventory)
    args = parser.parse_args(argv[1:])
    if args.version:
        print(version)
        sys.exit(0)
    loglevels = [logging.WARNING, logging.INFO, logging.DEBUG]
    if args.quiet and not args.log_file:
        args.verbose = 0
    loglevel = loglevels[min(len(loglevels)-1, args.verbose)]
    # always enable debug log if log file specificed
    if args.log_file:
        loglevel = logging.DEBUG
        log_handler = logging.handlers.WatchedFileHandler(args.log_file)
    else:
        log_handler = logging.StreamHandler()
    formatter = logging.Formatter('%(asctime)s %(levelname)s: %(module)s: '
                                  '%(funcName)s(): %(message)s')
    log_handler.setFormatter(formatter)
    logger = logging.getLogger(project_name)
    logger.addHandler(log_handler)
    logger.setLevel(loglevel)
    conf = inventory.NodeManager.load_conf(args.config)
    inventory.check_args(args, conf)
    if args.put or args.command or args.script or args.get:
        conf['shell_mode'] = True
        conf['do_print_results'] = True
    if args.no_clean:
        conf['clean'] = False
    if args.rqfile:
        conf['rqfile'] = []
        for file in args.rqfile:
            conf['rqfile'].append({'file': file, 'default': False})
    if args.days:
        conf['logs_days'] = args.days
    if args.logs_no_default:
        conf['logs_no_default'] = True
        args.logs = True
    if args.logs_speed or args.logs_speed_auto:
        conf['logs_speed_limit'] = True
    if args.logs_speed:
        conf['logs_speed'] = abs(args.logs_speed)
    if args.logs_coeff:
        conf['logs_size_coefficient'] = args.logs_coeff
    if conf['shell_mode']:
        filter = conf['hard_filter']
        # config cleanup for shell mode
        for k in Node.conf_actionable:
            conf[k] = [] if k in Node.conf_appendable else None
        for k in conf:
            if k.startswith(Node.conf_match_prefix):
                conf.pop(k)
        if args.put:
            conf[Node.pkey] = args.put
        if args.command:
            i = 0
            pad = str(len(str(len(args.command))))
            template = 'timmy_shell_mode_cmd_%0' + pad + 'd'
            for c in args.command:
                cmdname = template % i
                conf[Node.ckey].append({cmdname: c})
                i += 1
        if args.script:
            conf[Node.skey] = args.script
        if args.get:
            conf[Node.fkey] = args.get
    else:
        filter = conf['soft_filter']
    if args.get_logs:
        # this code should be after 'shell_mode' which cleans logs too
        args.logs = True
        for logs in args.get_logs:
            logs_conf = {}
            logs_conf['path'] = logs[0]
            if logs[1]:
                logs_conf['include'] = [logs[1]]
            if logs[2]:
                logs_conf['exclude'] = [logs[2]]
            if args.days:
                logs_conf['start'] = args.days
            conf['logs'].append(logs_conf)
    if args.role:
        filter['roles'] = args.role
    if args.env is not None:
        filter['cluster'] = [args.env]
    if args.id:
        filter['id'] = args.id
    if args.outputs_timestamp:
        conf['outputs_timestamp'] = True
    if args.dir_timestamp:
        conf['dir_timestamp'] = True
    if args.dest_file:
        conf['archive_dir'] = os.path.split(args.dest_file)[0]
        if not conf['archive_dir']:
            # this is mainly to see the path in logs instad of ""
            conf['archive_dir'] = os.getcwd()
        conf['archive_name'] = os.path.split(args.dest_file)[1]
    logger.info('Using rqdir: %s, rqfile: %s' %
                (conf['rqdir'], conf['rqfile']))
    nm = pretty_run(args.quiet, 'Initializing node data',
                    inventory.NodeManager,
                    kwargs={'conf': conf,
                            'nodes_json': args.nodes_json})
    if args.only_logs or args.logs:
        size = pretty_run(args.quiet, 'Calculating logs size',
                          nm.calculate_log_size, args=(args.maxthreads,))
        if size == 0:
            logger.warning('Size zero - no logs to collect.')
            has_logs = False
        else:
            has_logs = True
            print(('Total logs size to collect before compression: %dMB.\n'
                   'Compressed logs will take less space.') % (size/1048576))
            print('Log collection period: %s days.' % conf['logs_days'])
            enough_space = pretty_run(args.quiet, 'Checking free space',
                                      nm.is_enough_space)
            if not enough_space:
                logger.error('Not enough space for logs in "%s", exiting.' %
                             nm.conf['archive_dir'])
                sys.exit(100)
    if not args.only_logs:
        if nm.has(Node.pkey):
            pretty_run(args.quiet, 'Uploading files', nm.put_files)
        if nm.has(Node.ckey, Node.skey):
            pretty_run(args.quiet, 'Executing commands and scripts',
                       nm.run_commands, args=(args.maxthreads,))
        if nm.has('scripts_all_pairs'):
            pretty_run(args.quiet, 'Executing paired scripts',
                       nm.run_scripts_all_pairs, args=(args.maxthreads,))
        if nm.has(Node.fkey, Node.flkey):
            pretty_run(args.quiet, 'Collecting files and filelists',
                       nm.get_files, args=(args.maxthreads,))
        if not args.no_archive and nm.has(*Node.conf_archive_general):
            pretty_run(args.quiet, 'Creating outputs and files archive',
                       nm.create_archive_general, args=(60,))
    if (args.only_logs or args.logs) and has_logs and enough_space:
        msg = 'Collecting and packing logs'
        pretty_run(args.quiet, msg, nm.get_logs,
                   args=(conf['compress_timeout'],),
                   kwargs={'maxthreads': args.logs_maxthreads,
                           'fake': args.fake_logs})
    logger.info("Nodes:\n%s" % nm)
    if not args.quiet:
        print('Run complete. Node information:')
        print(nm)
    if conf['do_print_results']:
        if nm.has(Node.ckey, Node.skey):
            if not args.quiet:
                print('Results:')
            output_dict = {}
            maxlength = 0
            for node in nm.nodes.values():
                name, output = node.get_results(node.mapcmds)
                output_dict[node.ip] = {'name': name, 'output': output}
                name, output2 = node.get_results(node.mapscr)
                output_dict[node.ip]['output'] += output2
                maxlength = len(name) if len(name) > maxlength else maxlength
            for node in nm.sorted_nodes():
                for line in output_dict[node.ip]['output']:
                    name = output_dict[node.ip]['name'].rjust(maxlength)
                    print("%s: %s" % (name, line))
    if nm.has(Node.ckey, Node.skey, Node.fkey, Node.flkey) and not args.quiet:
        print('Outputs and/or files available in "%s".' % nm.conf['outdir'])
    if all([not args.no_archive, nm.has(*Node.conf_archive_general),
            not args.quiet]):
        print('Archives available in "%s".' % nm.conf['archive_dir'])

if __name__ == '__main__':
    sys.exit(main(sys.argv))



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

#    Copyright 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from setuptools import setup
from timmy.env import project_name, version
import sys

pname = project_name

if sys.argv[1] == 'test':
    setup_requires = ['pytest-runner']
else:
    setup_requires = []


setup(name=pname,
      version=version,
      author="Aleksandr Dobdin",
      author_email='dobdin@gmail.com',
      license='Apache2',
      url='https://github.com/adobdin/timmy',
      description=('Mirantis OpenStack Ansible-like tool for parallel node '
                   'operations: two-way data transfer, log collection, '
                   'remote command execution'),
      long_description=open('README.md').read(),
      packages=[pname, '%s.modules' % pname, '%s_data' % pname],
      install_requires=['pyyaml'],
      include_package_data=True,
      entry_points={'console_scripts': ['%s=%s.cli:main' % (pname, pname)]},
      setup_requires=setup_requires,
      tests_require=['pytest']
      )



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

#    Copyright 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tools import load_yaml_file
from tempfile import gettempdir
import os
from pkg_resources import resource_filename as resource_fn
from timmy.env import project_name


def init_default_conf():
    """Configuration parameters"""
    conf = {}
    conf['hard_filter'] = {}
    conf['soft_filter'] = {}
    conf['ssh_opts'] = ['-oConnectTimeout=2', '-oStrictHostKeyChecking=no',
                        '-oUserKnownHostsFile=/dev/null', '-oLogLevel=error',
                        '-oBatchMode=yes', '-oUser=root']
    conf['rsync_opts'] = ['-avzP', '--delete-before']
    conf['env_vars'] = ['OPENRC=/root/openrc', 'LC_ALL="C"', 'LANG="C"']
    conf['timeout'] = 30
    conf['prefix'] = 'nice -n 19 ionice -c 3'
    rqdir = 'rq'
    rqfile = 'default.yaml'
    data_package = '%s_data' % project_name
    conf['rqdir'] = os.path.join(resource_fn(data_package, rqdir))
    conf['rqfile'] = [{'file': os.path.join(conf['rqdir'], rqfile),
                      'default': True}]
    conf['compress_timeout'] = 3600
    conf['outdir'] = os.path.join(gettempdir(), 'timmy', 'info')
    conf['archive_dir'] = os.path.join(gettempdir(), 'timmy', 'archives')
    conf['archive_name'] = 'general.tar.gz'
    conf['outputs_timestamp'] = False
    conf['dir_timestamp'] = False
    conf['put'] = []
    conf['cmds'] = []
    conf['scripts'] = []
    conf['files'] = []
    conf['filelists'] = []
    conf['logs'] = []
    conf['logs_no_default'] = False  # skip logs defined in default.yaml
    conf['logs_days'] = 30
    conf['logs_speed_limit'] = False  # enable speed limiting of log transfers
    conf['logs_speed_default'] = 100  # Mbit/s, used when autodetect fails
    conf['logs_speed'] = 0  # To manually specify max bandwidth in Mbit/s
    conf['logs_size_coefficient'] = 1.05  # estimated logs compression ratio
    '''Shell mode - only run what was specified via command line.
    Skip actionable conf fields (see timmy/nodes.py -> Node.conf_actionable);
    Skip rqfile import;
    Skip any overrides (see Node.conf_match_prefix);
    Skip 'once' overrides (see Node.conf_once_prefix);
    Skip Fuel node;
    Print command execution results. Files and outputs will also be in a
    place specified by conf['outdir'], archive will also be created and put
    in a place specified by conf['archive_dir'].'''
    conf['shell_mode'] = False
    '''Print output of commands and scripts to stdout'''
    conf['do_print_results'] = False
    '''Clean - erase previous results in outdir and archive_dir dir, if any.'''
    conf['clean'] = True
    return conf


def update_conf(conf, filename):
    if filename is not None:
        conf_extra = load_yaml_file(filename)
        conf.update(**conf_extra)
    return conf


if __name__ == '__main__':
    import yaml
    conf = init_default_conf()
    conf = update_conf(conf, 'config.yaml')
    print(yaml.dump(conf))



#!/usr/bin/env python2
# -*- coding: utf-8 -*-

#    Copyright 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
tools module
"""

from flock import FLock
from pipes import quote
from tempfile import gettempdir
from timmy.env import project_name
import Queue
import json
import logging
import multiprocessing as mp
import os
import signal
import subprocess
import sys
import threading
import traceback
import yaml

logger = logging.getLogger(project_name)
slowpipe = '''
import sys
import time
while 1:
    a = sys.stdin.read(int(1250*%s))
    if a:
        sys.stdout.write(a)
        time.sleep(0.01)
    else:
        break
'''


def signal_wrapper(f):
    def wrapper(*args, **kwargs):
        setup_handle_sig()
        try:
            f(*args, **kwargs)
        except Exception as e:
            if not logger.handlers:
                logging.basicConfig()
            logger.error('Error: %s' % e, exc_info=True)
            for k in dir(e):
                '''debug: print all exception attrs except internal
                and except 'message', which is deprecated since Python 2.6'''
                if not k.startswith('__') and k != 'message':
                    v = getattr(e, k)
                    logger.debug('Error details: %s = %s' % (k, v))
            sys.exit(113)
    return wrapper


def handle_sig_usr1(sig, frame):
    debug_msg = 'Pid %d received USR1, printing current stack:'
    logger.critical(debug_msg % os.getpid())
    for line in traceback.format_stack(frame):
        for subline in line.splitlines():
            logger.critical(subline.rstrip())


def main_handle_sig(sig, frame):
    sig_msg = 'main application received signal %d, sending to children'
    logger.warning(sig_msg % sig)
    pids, errs, code = launch_cmd('pgrep -P %d' % os.getpid(), 2)
    if pids:
        for pid in pids.splitlines():
            try:
                os.kill(int(pid), sig)
            except OSError:
                pass
    sys.exit(sig)


def sub_handle_sig(sig, frame):
    sig_msg = 'subprocess received signal %d, sending to children, pid: %d'
    logger.warning(sig_msg % (sig, os.getpid()))
    signal.signal(sig, signal.SIG_IGN)
    try:
        os.killpg(os.getpid(), sig)
    except OSError:
        pass
    sys.exit(sig)


def setup_handle_sig(subprocess=False):
    if os.getpid() != os.getpgrp():
        os.setpgrp()
    sig_handler = main_handle_sig if not subprocess else sub_handle_sig
    for sig in [signal.SIGINT, signal.SIGTERM, signal.SIGHUP]:
        signal.signal(sig, sig_handler)
    signal.signal(signal.SIGUSR1, handle_sig_usr1)


def run_with_lock(f):
    def wrapper(*args, **kwargs):
        lock = FLock(os.path.join(gettempdir(), 'timmy_%s.lock' % f.__name__))
        if not lock.lock():
            logger.warning('Unable to obtain lock, skipping "%s"' %
                           f.__name__)
            return ''
        f(*args, **kwargs)
        lock.unlock()
    return wrapper


class RunItem():
    def __init__(self, target, args=None, key=None, logger=None):
        self.target = target
        self.args = args
        self.key = key
        self.process = None
        self.queue = None
        self.logger = logger or logging.getLogger(project_name)


class SemaphoreProcess(mp.Process):
    def __init__(self, semaphore, target, args=None, queue=None, logger=None):
        mp.Process.__init__(self)
        self.logger = logger or logging.getLogger(project_name)
        self.semaphore = semaphore
        self.target = target
        if not args:
            args = {}
        self.args = args
        self.queue = queue

    def run(self):
        setup_handle_sig(subprocess=True)
        fin_msg = 'finished subprocess, pid: %s'
        sem_msg = 'semaphore released by subprocess, pid: %s'
        try:
            result = self.target(**self.args)
            if self.queue:
                self.queue.put_nowait(result)
        except Exception as error:
            if self.queue:
                self.queue.put_nowait(error)
                self.queue.put_nowait(traceback.format_exc())
        finally:
            self.logger.debug(fin_msg % self.pid)
            self.semaphore.release()
            self.logger.debug(sem_msg % self.pid)


def run_batch(item_list, maxthreads, dict_result=False):
    exc_msg = 'exception in subprocess, pid: %s, details:'
    rem_msg = 'removing reference to finished subprocess, pid: %s'
    emp_msg = 'subprocess did not return results, pid: %s'

    def cleanup(launched):
        logger.info('cleaning up running subprocesses')
        for proc in launched.values():
            logger.debug('terminating subprocess, pid: %s' % proc.pid)
            proc.terminate()
            proc.join()

    def get_result(proc, key, results):
        try:
            results[key] = proc.queue.get(block=False)
        except Queue.Empty:
            if not proc.is_alive():
                logger.warning(emp_msg % proc.pid)

    def collect_results(l, join=False):
        results = {}
        remove_procs = []
        for key, proc in l.items():
            if not proc.is_alive() or join:
                logger.debug('joining subprocess, pid: %s' % proc.pid)
                while proc.is_alive():
                    get_result(proc, key, results)
                    proc.join(timeout=1)
                if key not in results:
                    get_result(proc, key, results)
                if key in results and isinstance(results[key], Exception):
                    exc_text = proc.queue.get()
                    logger.critical(exc_msg % proc.pid)
                    for line in exc_text.splitlines():
                        logger.critical('____%s' % line)
                    cleanup(l)
                    sys.exit(109)
                remove_procs.append(key)
        for key in remove_procs:
            logger.debug(rem_msg % key)
            l.pop(key)
        return results

    semaphore = mp.BoundedSemaphore(maxthreads)
    launched = {}
    results = {}
    if not dict_result:
        key = 0
    for run_item in item_list:
        results.update(collect_results(launched))
        semaphore.acquire(block=True)
        p = SemaphoreProcess(target=run_item.target,
                             semaphore=semaphore,
                             args=run_item.args,
                             queue=mp.Queue())
        p.start()
        if dict_result:
            launched[run_item.key] = p
            logger.debug('started subprocess, pid: %s, func: %s, key: %s' %
                         (p.pid, run_item.target, run_item.key))
        else:
            launched[key] = p
            key += 1
            logger.debug('started subprocess, pid:%s, func:%s, key:%s' %
                         (p.pid, run_item.target, key))

    results.update(collect_results(launched, True))
    if dict_result:
        return results
    else:
        return results.values()


def load_json_file(filename):
    """
    Loads json data from file
    """
    try:
        with open(filename, 'r') as f:
            return json.load(f)
    except IOError as e:
        logger.critical("I/O error(%s): file: %s; msg: %s" %
                        (e.errno, e.filename, e.strerror))
        sys.exit(107)
    except ValueError:
        logger.critical("Could not convert data", exc_info=True)
        sys.exit(108)


def load_yaml_file(filename):
    """
    Loads yaml data from file
    """
    try:
        with open(filename, 'r') as f:
            return yaml.load(f)
    except IOError as e:
        logger.critical("I/O error(%s): file: %s; msg: %s" %
                        (e.errno, e.filename, e.strerror))
        sys.exit(102)
    except ValueError:
        logger.critical("Could not convert data", exc_info=True)
        sys.exit(103)
    except yaml.parser.ParserError as e:
        logger.critical("Could not parse %s:\n%s" %
                        (filename, str(e)))
        sys.exit(105)


def mdir(directory):
    """
    Creates a directory if it doesn't exist
    """
    if directory and not os.path.exists(directory):
        logger.debug('creating directory %s' % directory)
        try:
            os.makedirs(directory)
        except OSError:
            logger.critical("Can't create a directory: %s" % directory)
            sys.exit(110)


def launch_cmd(cmd, timeout, input=None, ok_codes=None, decode=True):
    def _timeout_terminate(pid):
        try:
            os.kill(pid, 15)
            logger.error("launch_cmd: pid %d killed by timeout" % pid)
        except OSError:
            pass

    logger.debug('cmd: %s' % cmd)
    p = subprocess.Popen(cmd,
                         shell=True,
                         stdin=subprocess.PIPE,
                         stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE, close_fds=True)
    timeout_killer = None
    outs = None
    errs = None
    try:
        timeout_killer = threading.Timer(timeout, _timeout_terminate, [p.pid])
        timeout_killer.start()
        outs, errs = p.communicate(input=input)
        errs = errs.rstrip('\n')
        if decode:
            outs = outs.decode('utf-8')
            errs = errs.decode('utf-8')
    finally:
        if timeout_killer:
            timeout_killer.cancel()
        input = input.decode('utf-8') if input else None
        logger.debug(('___command: %s\n'
                      '_______pid: %s\n'
                      '_exit_code: %s\n'
                      '_____stdin: %s\n'
                      '____stderr: %s') % (cmd, p.pid, p.returncode, input,
                                           errs))
    return outs, errs, p.returncode


def ssh_node(ip, command='', ssh_opts=None, env_vars=None, timeout=15,
             filename=None, inputfile=None, outputfile=None,
             ok_codes=None, input=None, prefix=None, decode=True):
    if not ssh_opts:
        ssh_opts = ''
    if not env_vars:
        env_vars = ''
    if type(ssh_opts) is list:
        ssh_opts = ' '.join(ssh_opts)
    if type(env_vars) is list:
        env_vars = ' '.join(env_vars)
    if (ip in ['localhost', '127.0.0.1']) or ip.startswith('127.'):
        logger.debug("skip ssh")
        bstr = "%s timeout '%s' bash -c " % (
               env_vars, timeout)
    else:
        bstr = "timeout '%s' ssh -T %s '%s' '%s' " % (
               timeout, ssh_opts, ip, env_vars)
    if filename is None:
        cmd = '%s %s' % (bstr, quote(prefix + ' ' + command))
        if inputfile is not None:
            '''inputfile and stdin will not work together,
            give priority to inputfile'''
            input = None
            cmd = "%s < '%s'" % (cmd, inputfile)
    else:
        cmd = "%s'%s bash -s' < '%s'" % (bstr, prefix, filename)
    if outputfile is not None:
        cmd = "%s > '%s'" % (cmd, outputfile)
    logger.info("cmd: %s" % cmd)
    cmd = ("input=\"$(cat | xxd -p)\"; trap 'kill $pid' 15; " +
           "trap 'kill $pid' 2; echo -n \"$input\" | xxd -r -p | " + cmd +
           ' &:; pid=$!; wait $!')
    return launch_cmd(cmd, timeout, input=input,
                      ok_codes=ok_codes, decode=decode)


def get_files_rsync(ip, data, ssh_opts, rsync_opts, dpath, timeout=15):
    if type(ssh_opts) is list:
        ssh_opts = ' '.join(ssh_opts)
    if type(rsync_opts) is list:
        rsync_opts = ' '.join(rsync_opts)
    if (ip in ['localhost', '127.0.0.1']) or ip.startswith('127.'):
        logger.info("skip ssh rsync")
        cmd = ("timeout '%s' rsync %s --include-from=- / '%s' --exclude='*'" %
               (timeout, rsync_opts, dpath))
    else:
        cmd = ("timeout '%s' rsync %s -e 'ssh %s -oCompression=no' "
               "--include-from=- '%s':/ '%s' --exclude='*'" %
               (timeout, rsync_opts, ssh_opts, ip, dpath))
    logger.debug("command:%s\ndata:\n%s" % (cmd, data))
    if data == '':
        return cmd, '', 127
    return launch_cmd(cmd, timeout, input=data)


def get_file_scp(ip, file, ddir, ssh_opts, timeout=600, recursive=False):
    if type(ssh_opts) is list:
        ssh_opts = ' '.join(ssh_opts)
    dest = os.path.split(os.path.normpath(file).lstrip(os.path.sep))[0]
    ddir = os.path.join(os.path.normpath(ddir), dest)
    mdir(ddir)
    r = '-r ' if recursive else ''
    cmd = ("timeout '%s' scp %s -p -q %s'%s':'%s' '%s'" %
           (timeout, ssh_opts, r, ip, file, ddir))
    return launch_cmd(cmd, timeout)


def put_file_scp(ip, file, dest, ssh_opts, timeout=600, recursive=True):
    if type(ssh_opts) is list:
        ssh_opts = ' '.join(ssh_opts)
    r = '-r ' if recursive else ''
    cmd = ("timeout '%s' scp %s -p -q %s'%s' '%s':'%s'" %
           (timeout, ssh_opts, r, file, ip, dest))
    return launch_cmd(cmd, timeout)


def free_space(destdir, timeout):
    cmd = ("df %s --block-size K 2> /dev/null"
           " | tail -n 1 | awk '{print $4}' | sed 's/K//g'") % (destdir)
    return launch_cmd(cmd, timeout)


# wrap non-list into list
def w_list(value):
    return value if type(value) == list else [value]


def all_pairs(items):
    def incomplete(i_set, p_dict):
        for i, p_set in p_dict.items():
            not_paired = i_set.difference(p_set).difference([i])
            if not_paired:
                return not_paired

    items_set = set(items)
    pairs = []
    paired = {}
    for i in items_set:
        paired[i] = set()
    while incomplete(items_set, paired):
        busy = set()
        current_pairs = []
        for i in [i for i in items if items_set.difference(paired[i])]:
            can_pair = incomplete(items_set.difference(busy), {i: paired[i]})
            if i not in busy and can_pair:
                pair_i = next(iter(can_pair))
                current_pairs.append([i, pair_i])
                busy.add(i)
                busy.add(pair_i)
                paired[i].add(pair_i)
        pairs.append(current_pairs)
    return pairs


if __name__ == '__main__':
    sys.exit(0)
